from __future__ import annotations

from pydantic import BaseModel, Field

from .field import BaseExtraField


class Device[ExtraT: BaseExtraField](BaseModel):
    system: str
    hardware: str
    group: str
    main: bool = False
    desc: str = Field("")

    extra: ExtraT = Field(default_factory=BaseExtraField)

    def get_name(self):
        return f"{self.group}-{self.hardware}-{self.system}"

    def get_domain(self):
        return f"{self.hardware}-{self.system}.ssh.{self.group}"



from __future__ import annotations

from typing import Sequence

from ..service import Service
from .base import GeneratorBase


class CaddyFileGenerator(GeneratorBase):
    def __init__(self, services: Sequence[Service]):
        super().__init__()
        self.services = services

    def generate(self):
        for service in self.services:
            self.add_block(f"{service.get_domain()}.bone6.top {{", "}", indentation=4)
            self.add_line(f"reverse_proxy {service.target}")
            self.add_line("tls {")
            self.add_line("    dns dnspod {$DNSPOD_ID},{$DNSPOD_TOKEN}")
            self.add_line("}")
        super().generate()



from __future__ import annotations

from pydantic import BaseModel


class Service(BaseModel):
    name: str
    group: str
    target: str

    def get_domain(self):
        return f"{self.name}.{self.group}"



from __future__ import annotations

import json
from typing import Sequence

from ..device import Device
from ..field import DefaultExtraField
from .base import GeneratorBase


class DNSConfigGenerator(GeneratorBase):
    def __init__(self, devices: Sequence[Device[DefaultExtraField]]):
        super().__init__()
        self.devices = devices

    def generate(self):
        record_list = []
        for device in self.devices:
            name = device.get_domain()
            if device.extra.dns is not None:
                target = device.extra.dns.target
            else:
                target = "unknown"
            if device.main:
                record_list.append((device.group, f"{name}.bone6.top"))
            record_list.append((f"{name}", target))

        self._generated_code = json.dumps({"domain": "bone6.top", "records": record_list}, indent=2)



from __future__ import annotations

from typing import Sequence

from loguru import logger

from ..device import Device
from ..field import DefaultExtraField
from .base import GeneratorBase


class SSHHostsGenerator(GeneratorBase):
    def __init__(self, devices: Sequence[Device[DefaultExtraField]]):
        super().__init__()
        self.devices = devices

    def generate(self):
        for device in self.devices:
            if device.extra.ssh is None:
                logger.warning(f"device {device} has no ssh config")
                continue
            self.add_line(f"Host {device.get_name()}")
            self.add_line(f"  HostName {device.get_domain()}.bone6.top")
            self.add_line(f"  Port {device.extra.ssh.port}")
            self.add_line(f"  User {device.extra.ssh.user}")
            self.add_line(f"  ForwardAgent {"yes" if device.extra.ssh.forward_agent else "no"}")
            self.add_line(f"  #_Desc {device.desc}")
            self.add_line()

        super().generate()



from __future__ import annotations

import os

from loguru import logger


class Block:
    def __init__(self, start_line: str = "", end_line: str = "", *, indentation: int = 0):
        self.start_line = start_line
        self.end_line = end_line
        self.indentation = indentation
        self._lines: list[str] = []

    def add_line(self, line: str = ""):
        self._lines.append(f"{' '*self.indentation}{line}")

    def generate(self):
        return "\n".join([self.start_line, *self._lines, self.end_line])


class GeneratorBase:
    def __init__(self):
        self._generated_code: str | None = None
        self._cur_block = Block("", "")
        self._block: list[Block] = [self._cur_block]

    def add_block(self, start_line: str = "", end_line: str = "", *, indentation: int = 0):
        self._cur_block = Block(start_line, end_line, indentation=indentation)
        self._block.append(self._cur_block)
        # return

    def add_line(self, line: str = ""):
        self._cur_block.add_line(line)

    def generate(self):
        self._generated_code = "\n".join(block.generate() for block in self._block)

    def clear(self):
        self._generated_code = None

    @property
    def code(self):
        if self._generated_code is None:
            self.generate()
            assert self._generated_code is not None, "generate() should set self._generated_code"

        return self._generated_code

    def write(self, path: str):
        path = os.path.expanduser(path)
        path = os.path.expandvars(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        logger.info(f"writed to {path}")
        open(path, "w").write(self.code)



from __future__ import annotations

from sys import stdout

from loguru import logger

from auto_config.device import Device

logger.remove()
logger.add(stdout, level="DEBUG")


def test_get_name():
    device = Device(system="system1", hardware="hardware1", group="group1", main=True, desc="")
    assert device.get_name() == "group1-hardware1-system1"


def test_get_domain():
    device = Device(system="system1", hardware="hardware1", group="group1", main=True, desc="")
    assert device.get_domain() == "hardware1-system1.ssh.group1"



from __future__ import annotations

from .ansible_hosts import AnsibleHostsGenerator
from .base import GeneratorBase
from .caddy_file import CaddyFileGenerator
from .dns_config import DNSConfigGenerator
from .ssh_hosts import SSHHostsGenerator
